// Copyright (C) 2011  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.
#ifndef DLIB_IMGLAB_CONVERT_PASCAl_XML_H__
#define DLIB_IMGLAB_CONVERT_PASCAl_XML_H__

#include "common.h"
#include <dlib/cmd_line_parser.h>

void convert_pascal_xml(const dlib::command_line_parser& parser);

#endif // DLIB_IMGLAB_CONVERT_PASCAl_XML_H__

