#ifndef VIENNA_RNA_PACKAGE_ENERGY_PARAMETER_SETS_H
#define VIENNA_RNA_PACKAGE_ENERGY_PARAMETER_SETS_H

/* The strings below are automatically generated from corresponding .par files */


static const unsigned char parameter_set_dna_mathews1999[] = {
#include "misc/dna_mathews1999.hex"
};

static const unsigned char parameter_set_dna_mathews2004[] = {
#include "misc/dna_mathews2004.hex"
};

static const unsigned char parameter_set_rna_andronescu2007[] = {
#include "misc/rna_andronescu2007.hex"
};

static const unsigned char parameter_set_rna_langdon2018[] = {
#include "misc/rna_langdon2018.hex"
};

static const unsigned char parameter_set_rna_misc_special_hairpins[] = {
#include "misc/rna_misc_special_hairpins.hex"
};

static const unsigned char parameter_set_rna_turner1999[] = {
#include "misc/rna_turner1999.hex"
};

static const unsigned char parameter_set_rna_turner2004[] = {
#include "misc/rna_turner2004.hex"
};

static const unsigned char parameter_set_rna_mod_7DA_parameters[] = {
#include "misc/rna_mod_7DA_parameters.hex"
};

static const unsigned char parameter_set_rna_mod_inosine_parameters[] = {
#include "misc/rna_mod_inosine_parameters.hex"
};

static const unsigned char parameter_set_rna_mod_m6A_parameters[] = {
#include "misc/rna_mod_m6A_parameters.hex"
};

static const unsigned char parameter_set_rna_mod_pseudouridine_parameters[] = {
#include "misc/rna_mod_pseudouridine_parameters.hex"
};

static const unsigned char parameter_set_rna_mod_purine_parameters[] = {
#include "misc/rna_mod_purine_parameters.hex"
};

static const unsigned char parameter_set_rna_mod_dihydrouridine_parameters[] = {
#include "misc/rna_mod_dihydrouridine_parameters.hex"
};


#endif
