// Copyright (C) 2015  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.
#ifndef DLIB_IMGLAB_ClUSTER_H_
#define DLIB_IMGLAB_ClUSTER_H_

#include <dlib/cmd_line_parser.h>

int cluster_dataset(const dlib::command_line_parser& parser);

#endif //DLIB_IMGLAB_ClUSTER_H_

