extern const char   wall;
extern const char   bg;
extern const char   star;
extern const char   probe;
extern const char   start[];
extern const char   end[];
extern const char   success[];
extern const char   injector[];
extern unsigned int injector_len;
extern const char   flash[];
extern const char   head1l[];
extern const char   head2l[];
extern const char   lvlstr[];
extern const char   inv[];
