#ifndef VIENNA_RNA_PACKAGE_SUBOPT_ZUKER_DEPRECATED_H
#define VIENNA_RNA_PACKAGE_SUBOPT_ZUKER_DEPRECATED_H

/**
 *  @file       subopt_zuker.h
 *  @brief      Deprecated include file for Zuker-style subopt API
 *  @deprecated Use ViennaRNA/subopt/zuker.h instead
 */

#ifndef VRNA_DISABLE_BACKWARD_COMPATIBILITY
# ifdef VRNA_WARN_DEPRECATED
#warning "Including deprecated header file <ViennaRNA/subopt_zuker.h>! Use <ViennaRNA/subopt/zuker.h> instead!"
# endif
#include <ViennaRNA/subopt/zuker.h>
#endif

#endif
