#ifndef DLIB_REVISION_H
// Version:  20.0
// Date:     Tue May 27 08:07:11 PM EDT 2025
// Git Changeset ID:  2e0b3846b4dcd0953951a9c0172503c666975229
#define DLIB_MAJOR_VERSION  20
#define DLIB_MINOR_VERSION  0
#define DLIB_PATCH_VERSION  0
#endif
