#ifndef VIENNA_RNA_PACKAGE_PLOT_SVG_TEMPLATES_H
#define VIENNA_RNA_PACKAGE_PLOT_SVG_TEMPLATES_H

/* The strings below are automatically generated from corresponding .svg files */


static const char SVG_structure_plot_header[] = {
#include "svg/structure_plot_header.hex"
};

static const char SVG_structure_plot_footer[] = {
#include "svg/structure_plot_footer.hex"
};


#endif
