#ifndef VIENNA_RNA_PACKAGE_DP_MATRICES_DEPRECATED_H
#define VIENNA_RNA_PACKAGE_DP_MATRICES_DEPRECATED_H

/**
 *  @file       dp_matrices.h
 *  @brief      Deprecated include file for DP matrices API
 *  @deprecated Use ViennaRNA/datastructures/dp_matrices.h instead
 */

#ifndef VRNA_DISABLE_BACKWARD_COMPATIBILITY
# ifdef VRNA_WARN_DEPRECATED
#warning "Including deprecated header file <ViennaRNA/dp_matrices.h>! Use <ViennaRNA/datastructures/dp_matrices.h> instead!"
# endif
#include <ViennaRNA/datastructures/dp_matrices.h>
#endif

#endif
