#ifndef VIENNA_RNA_PACKAGE_STRUCTURAL_DOMAINS_H
#define VIENNA_RNA_PACKAGE_STRUCTURAL_DOMAINS_H

/**
 *  @file structured_domains.h
 *  @ingroup domains_struc
 *
 *  @brief  This module provides interfaces that deal with additional structured domains in the folding grammar.
 */

typedef struct vrna_structured_domains_s vrna_sd_t;


struct vrna_structured_domains_s {
  char __placeholder; /* dummy placeholder to not leave this struct empty */
};

#endif
