#ifndef VIENNA_RNA_PACKAGE_PROBING_DATA_PRIORS_H
#define VIENNA_RNA_PACKAGE_PROBING_DATA_PRIORS_H

/* The strings below are automatically generated from corresponding .data files */


static const char probing_data_prior_probing_1M7_paired[] = {
#include "misc/prior_probing_1M7_paired.hex"
};

static const char probing_data_prior_probing_1M7_unpaired[] = {
#include "misc/prior_probing_1M7_unpaired.hex"
};

static const char probing_data_prior_probing_DMS_paired[] = {
#include "misc/prior_probing_DMS_paired.hex"
};

static const char probing_data_prior_probing_DMS_unpaired[] = {
#include "misc/prior_probing_DMS_unpaired.hex"
};


#endif
