// Copyright (C) 2011  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.
#ifndef DLIB_IMGLAB_FLIP_DaTASET_H__
#define DLIB_IMGLAB_FLIP_DaTASET_H__


#include <dlib/cmd_line_parser.h>

void flip_dataset(const dlib::command_line_parser& parser);

#endif // DLIB_IMGLAB_FLIP_DaTASET_H__

