// Copyright (C) 2009  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.
#ifdef DLIB_ALL_SOURCE_END
#include "dlib_basic_cpp_build_tutorial.txt"
#endif

#ifndef DLIB_OPEnCV_HEADER
#define DLIB_OPEnCV_HEADER

#include "opencv/cv_image.h"
#include "opencv/to_open_cv.h"

#endif // DLIB_OPEnCV_HEADER




