#ifndef VIENNA_RNA_PACKAGE_PLOT_PS_TEMPLATES_H
#define VIENNA_RNA_PACKAGE_PLOT_PS_TEMPLATES_H

/* The strings below are automatically generated from corresponding .ps files */


static const unsigned char PS_structure_plot_macro_base[] = {
#include "postscript/structure_plot_macro_base.hex"
};

static const unsigned char PS_structure_plot_macro_extras[] = {
#include "postscript/structure_plot_macro_extras.hex"
};

static const unsigned char PS_dot_plot_macro_base[] = {
#include "postscript/dot_plot_macro_base.hex"
};

static const unsigned char PS_dot_plot_macro_turn[] = {
#include "postscript/dot_plot_macro_turn.hex"
};

static const unsigned char PS_dot_plot_macro_sd[] = {
#include "postscript/dot_plot_macro_sd.hex"
};

static const unsigned char PS_dot_plot_macro_ud[] = {
#include "postscript/dot_plot_macro_ud.hex"
};

static const unsigned char PS_dot_plot_macro_sc_motifs[] = {
#include "postscript/dot_plot_macro_sc_motifs.hex"
};

static const unsigned char PS_dot_plot_macro_linear_data[] = {
#include "postscript/dot_plot_macro_linear_data.hex"
};

static const unsigned char PS_aln_macro_base[] = {
#include "postscript/aln_macro_base.hex"
};


#endif
