// Copyright (C) 2011  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.
#ifndef DLIB_LSh_
#define DLIB_LSh_


#include "lsh/projection_hash.h"
#include "lsh/create_random_projection_hash.h"
#include "lsh/hashes.h"


#endif // DLIB_LSh_


